"""*EPUB handling module*."""

import io
import shutil
import tempfile
from pathlib import Path
from typing import Any, List

import ebookmeta
from PIL import Image, ImageEnhance
import iscc_sdk as idk
import zipfile
import lxml
from loguru import logger as log


__all__ = [
    "epub_thumbnail",
    "epub_meta_embed",
    "epub_cover",
]


def epub_thumbnail(fp):
    # type: (str|Path) -> Image.Image
    """
    Creat thumbnail from EPUB document cover image.

    :param fp: Filepath to EPUB document.
    :return: Thumbnail image as PIL Image object
    """
    fp = Path(fp)
    data = epub_cover(fp)
    img = Image.open(io.BytesIO(data))
    size = idk.sdk_opts.image_thumbnail_size
    img.thumbnail((size, size), resample=idk.LANCZOS)
    return ImageEnhance.Sharpness(img.convert("RGB")).enhance(1.4)


def epub_meta_embed(fp, meta):
    # type: (str|Path, idk.IsccMeta) -> str
    """
    Embed metadata into a copy of the EPUB file.

    :param fp: Filepath to source EPUB file
    :param IsccMeta meta: Metadata to embed into EPUB
    :return: Filepath to the new EPUB file with updated metadata
    """
    fp = Path(fp)
    tempdir = tempfile.mkdtemp()
    tempepub = shutil.copy(fp, tempdir)
    new_meta = ebookmeta.get_metadata(tempepub)
    if meta.name:
        new_meta.title = meta.name
    if meta.description:
        new_meta.description = meta.description
    if meta.creator:
        if isinstance(meta.creator, str):
            new_meta.set_author_list_from_string(meta.creator)
    ebookmeta.set_metadata(tempepub, new_meta)
    return tempepub


def epub_cover(fp):  # pragma: no cover
    # type: (str|Path) -> bytes
    """
    Extract the cover image bytes from an EPUB file.

    This function attempts to locate the cover image by first checking the metadata
    cover reference, then falling back to scanning for image files with 'cover' in the name and
    if that fails it returns the first image file from the manifest.
    The function also logs the relative path to the image within the epub which was identified as
    cover image. If no image is found, it raises an error.

    :param fp: Filepath to EPUB file
    :return: Raw bytes of the cover image
    :raises IsccExtractionError: If no cover image can be found.
    """
    fp = Path(fp)
    cover_path = None
    image_paths = []

    try:
        with zipfile.ZipFile(fp, "r") as archive:
            # Find OPF file path from container.xml
            container_path = "META-INF/container.xml"
            if container_path not in archive.namelist():
                raise idk.IsccExtractionError(f"Missing {container_path} in {fp}")

            container_xml = archive.read(container_path)
            container_root = lxml.etree.fromstring(container_xml)
            opf_path = container_root.xpath(
                "//o:rootfile/@full-path",
                namespaces={"o": "urn:oasis:names:tc:opendocument:xmlns:container"},
            )[0]

            # Parse OPF file
            opf_xml = archive.read(opf_path)
            opf_root = lxml.etree.fromstring(opf_xml)
            opf_ns = {"opf": "http://www.idpf.org/2007/opf"}

            # 1. Check metadata for cover reference
            cover_id = opf_root.xpath(
                "//opf:metadata/opf:meta[@name='cover']/@content", namespaces=opf_ns
            )
            if cover_id:
                cover_id = cover_id[0]
                cover_href = opf_root.xpath(
                    f"//opf:manifest/opf:item[@id='{cover_id}']/@href", namespaces=opf_ns
                )
                if cover_href:
                    cover_path = Path(opf_path).parent / cover_href[0]
                    log.debug(f"Found cover image via metadata: {cover_path.as_posix()}")

            # 2. Scan manifest for images with 'cover' in the name
            if not cover_path:
                manifest_items = opf_root.xpath("//opf:manifest/opf:item", namespaces=opf_ns)
                for item in manifest_items:
                    media_type = item.get("media-type", "")
                    href = item.get("href")
                    if media_type.startswith("image/") and href:
                        item_path = (Path(opf_path).parent / href).as_posix()
                        image_paths.append(item_path)
                        if "cover" in href.lower():
                            cover_path = Path(item_path)
                            log.debug(
                                f"Found cover image via manifest scan: {cover_path.as_posix()}"
                            )
                            break

            # 3. Fallback to the first image in the manifest
            if not cover_path and image_paths:
                cover_path = Path(image_paths[0])
                log.debug(f"Using first image from manifest as cover: {cover_path.as_posix()}")

            if not cover_path:
                raise idk.IsccExtractionError(f"No cover image found in {fp}")

            # Ensure the path is relative to the archive root
            cover_path_str = cover_path.as_posix()
            if cover_path_str not in archive.namelist():
                # Try resolving relative paths if needed (though Path should handle this)
                # This part might need adjustment based on EPUB structure variations
                log.warning(
                    f"Cover path {cover_path_str} not directly in archive, attempting resolution."
                )
                # Basic check if it exists at all
                found = False
                for name in archive.namelist():
                    if name.endswith(cover_path.name):
                        cover_path_str = name
                        found = True
                        log.debug(f"Resolved cover path to {cover_path_str}")
                        break
                if not found:
                    raise idk.IsccExtractionError(
                        f"Cover image path {cover_path_str} not found in archive {fp}"
                    )

            # Extract image bytes
            return archive.read(cover_path_str)

    except zipfile.BadZipFile:
        raise idk.IsccExtractionError(f"Invalid EPUB (Zip) file: {fp}")
    except (lxml.etree.XMLSyntaxError, IndexError) as e:
        raise idk.IsccExtractionError(f"Failed to parse EPUB metadata for {fp}: {e}")
    except KeyError as e:
        raise idk.IsccExtractionError(f"Cover image file {e} not found within EPUB archive {fp}")


def epub_process_container(fp, **options):
    # type: (str|Path, Any) -> List[idk.IsccMeta]
    """
    Extract and process images from EPUB file.
    Skips processing for fixed layout EPUBs.

    :param fp: Filepath to EPUB file
    :param options: Processing options
    :return: List of IsccMeta objects for embedded images
    """
    fp = Path(fp)
    parts = []

    # Check if the EPUB is fixed layout
    if is_fixed_layout_epub(fp):  # pragma: no cover
        log.info(f"Skipping container processing for fixed layout EPUB: {fp.name}")
        return parts

    with tempfile.TemporaryDirectory() as temp_dir:
        # Extract images from EPUB
        images = epub_extract_images(fp, temp_dir)

        # Process each image
        for img_path in images:
            try:
                # Generate ISCC for each image
                img_meta = idk.code_iscc(img_path, **options)
                parts.append(img_meta.dict())
            except Exception as e:  # pragma: no cover
                log.warning(f"Failed to process embedded image {img_path.name}: {e}")

    return parts


def epub_extract_images(fp, output_dir):
    # type: (str|Path, str|Path) -> List[Path]
    """
    Extract images from EPUB file to output directory where both
    width and height are >= min_image_size.

    :param fp: Filepath to EPUB file
    :param output_dir: Directory to extract images to
    :return: List of paths to extracted images
    """
    fp = Path(fp)
    output_dir = Path(output_dir)
    extracted_images = []
    min_size = idk.sdk_opts.min_image_size

    with zipfile.ZipFile(fp, "r") as archive:
        # Identify image files in the archive
        for item in archive.namelist():
            if item.lower().endswith((".jpg", ".jpeg", ".png", ".gif", ".webp")):
                # Extract the image
                img_data = archive.read(item)

                # Check image dimensions before extracting
                try:
                    img = Image.open(io.BytesIO(img_data))
                    width, height = img.size

                    # Only extract if both dimensions are >= min_size
                    if width >= min_size and height >= min_size:
                        img_filename = Path(item).name
                        img_path = output_dir / img_filename

                        # Write to temp file
                        with open(img_path, "wb") as img_file:
                            img_file.write(img_data)

                        extracted_images.append(img_path)
                except Exception as e:  # pragma: no cover
                    log.warning(f"Failed to process image {item} from EPUB: {e}")

    return extracted_images


def is_fixed_layout_epub(fp):  # pragma: no cover
    # type: (str|Path) -> bool
    """
    Check if an EPUB is a fixed layout publication.

    :param fp: Path to EPUB file
    :return: True if the EPUB is fixed layout, False otherwise
    """
    fp = Path(fp)

    try:
        with zipfile.ZipFile(fp, "r") as archive:
            # Find OPF file path from container.xml
            container_path = "META-INF/container.xml"
            if container_path not in archive.namelist():
                log.warning(f"Missing {container_path} in {fp}")
                return False

            container_xml = archive.read(container_path)
            container_root = lxml.etree.fromstring(container_xml)
            opf_path = container_root.xpath(
                "//o:rootfile/@full-path",
                namespaces={"o": "urn:oasis:names:tc:opendocument:xmlns:container"},
            )[0]

            # Parse OPF file
            opf_xml = archive.read(opf_path)
            opf_root = lxml.etree.fromstring(opf_xml)

            # Define necessary namespaces for XPath queries
            namespaces = {
                "opf": "http://www.idpf.org/2007/opf",
                "dc": "http://purl.org/dc/elements/1.1/",
            }

            # Check for fixed layout indicators in metadata
            # Method 1: EPUB 3.0 standard - using namespaces
            fixed_layout = opf_root.xpath(
                "//opf:metadata/opf:meta[@property='rendition:layout' and text()='pre-paginated']",
                namespaces=namespaces,
            )
            if fixed_layout:
                return True

            # Method 2: Alternative specification - using namespaces
            fixed_layout = opf_root.xpath(
                "//opf:metadata/opf:meta[@name='fixed-layout' and (@content='true' or @content='yes')]",
                namespaces=namespaces,
            )
            if fixed_layout:
                return True

            # Method 3: Some EPUBs might use meta elements without explicit namespace
            fixed_layout = opf_root.xpath(
                "//meta[@property='rendition:layout' and text()='pre-paginated']"
            )
            if fixed_layout:
                return True

            fixed_layout = opf_root.xpath(
                "//meta[@name='fixed-layout' and (@content='true' or @content='yes')]"
            )
            if fixed_layout:
                return True

            return False

    except Exception as e:
        log.warning(f"Error checking if EPUB is fixed layout: {e}")
        return False


# Register EPUB container processor
idk.register_container_processor("application/epub+zip", epub_process_container)



"""*SDK main top-level functions*."""

from typing import Any

from loguru import logger as log
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from PIL import Image
import iscc_core as ic
import iscc_sdk as idk


__all__ = [
    "code_iscc",
    "code_meta",
    "code_content",
    "code_text",
    "code_image",
    "code_audio",
    "code_video",
    "code_data",
    "code_instance",
]


def code_iscc(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate a complete ISCC-CODE for the given file.

    This function creates a full ISCC-CODE by combining Meta, Content, Data, and Instance Codes.
    It automatically detects the media type and processes the file accordingly.

    The function performs the following steps:
    1. Reads the file and determines its MIME type.
    2. Generates Instance and Data Codes for all file types.
    3. For supported media types, generates Content and Meta Codes.
    4. Combines all generated codes into a single ISCC-CODE.
    5. Merges metadata from all ISCC units.

    Note:

    - This function uses multithreading to improve performance.
    - The behavior can be customized through the `sdk_opts` settings. For example, setting
      `fallback` to True will allow processing of unsupported media types in a
      fallback mode instead of raising an exception.

    :param fp: str or Path object representing the filepath of the file to process.
    :key fallback: Process unsupported media types. Default: False
    :key add_units: Include ISCC-UNITS in metadata. Default: False
    :key create_meta: Create Meta-Code unit from embedded metadata. Default: True
    :key wide: Enable wide mode for ISCC-SUM with Data & Instance codes only. Default: False
    :return: IsccMeta object with complete ISCC-CODE and merged metadata from all ISCC-UNITs.
    :raises idk.IsccUnsupportedMediatype:
        If the media type is not supported. By default, the function will raise this exception for
        unsupported media types, as sdk_opts.fallback is False by default.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    iscc_meta = dict(filename=fp.name)
    iscc_units = []

    with open(fp, "rb") as infile:
        data = infile.read(4096)

    mime = idk.mediatype_guess(data, file_name=fp.name)
    iscc_meta["mediatype"] = mime

    with ThreadPoolExecutor() as executor:
        # Always process instance and data
        instance_future = executor.submit(code_instance, fp, **options)
        data_future = executor.submit(code_data, fp, **options)
        try:
            mode = idk.mediatype_to_mode(mime)
            log.debug(f"Processing {fp.name} - media type: {mime} - processing mode: {mode}")

        except idk.IsccUnsupportedMediatype:
            if not opts.fallback:
                raise
            log.warning(f"Processing {fp.name} - media type: {mime} - processing mode: sum")
        else:
            # Process content and meta for supported media types
            content_future = executor.submit(code_content, fp, **options)
            if opts.create_meta:
                meta_future = executor.submit(code_meta, fp, **options)
                meta = meta_future.result()
                iscc_units.append(meta.iscc)

            content = content_future.result()
            iscc_units.append(content.iscc)
            iscc_meta.update(content.dict())
            if opts.create_meta:
                iscc_meta.update(meta.dict())

        finally:
            # Wait for instance and data to complete
            instance = instance_future.result()
            data = data_future.result()
            iscc_units.extend([data.iscc, instance.iscc])
            iscc_meta.update(instance.dict())
            iscc_meta.update(data.dict())

    if opts.add_units:
        iscc_meta["units"] = iscc_units
    # Compose ISCC-CODE
    iscc_code = ic.gen_iscc_code_v0(iscc_units, wide=opts.wide)
    iscc_meta.update(iscc_code)
    iscc_meta["generator"] = f"iscc-sdk - v{idk.__version__}"

    result = idk.IsccMeta.construct(**iscc_meta)

    if opts.process_container:
        parts = idk.process_container(fp, **options)
        if parts:
            result.parts = parts

    return result


def code_meta(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Meta-Code from digital asset.

    Creates an ISCC Meta-Code based on normalized metadata extracted from the file.
    If no name is found in metadata, the filename will be used instead.

    :param fp: Filepath used for Meta-Code creation.
    :key bits: Bit-length of the generated Meta-Code UNIT. Default: 64
    :return: ISCC metadata including Meta-Code and extracted metadata fields.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)

    meta = idk.extract_metadata(fp).dict()

    # Pre-Check if we have a name after normalization, else use filename.
    name = meta.get("name")
    name = "" if name is None else name
    name = ic.text_clean(name)
    name = ic.text_remove_newlines(name)
    name = ic.text_trim(name, ic.core_opts.meta_trim_name)
    if not name:
        meta["name"] = idk.text_name_from_uri(fp)
        log.debug(f"Acquired Meta-Code `name` from filename: {meta['name']}")

    metacode = ic.gen_meta_code_v0(
        name=meta.get("name"),
        description=meta.get("description"),
        meta=meta.get("meta"),
        bits=opts.bits,
    )

    meta.update(metacode)
    return idk.IsccMeta.construct(**meta)


def code_content(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Detect mediatype and create corresponding Content-Code.

    Analyzes the file to determine its media type and routes the processing to the
    appropriate specialized function (code_text, code_image, code_audio, or code_video).

    :param fp: Filepath
    :key extract_meta: Whether to extract metadata. Default: True
    :key create_thumb: Whether to create a thumbnail. Default: True
    :return: Content-Code wrapped in ISCC metadata.
    :raises idk.IsccUnsupportedMediatype: If the media type is not supported.
    """
    fp = Path(fp)
    schema_org_map = {
        "text": "TextDigitalDocument",
        "image": "ImageObject",
        "audio": "AudioObject",
        "video": "VideoObject",
    }

    mediatype, mode = idk.mediatype_and_mode(fp)

    if mode == "image":
        cc = code_image(fp, **options)
    elif mode == "audio":
        cc = code_audio(fp, **options)
    elif mode == "video":
        cc = code_video(fp, **options)
    elif mode == "text":
        cc = code_text(fp, **options)
    else:  # pragma nocover
        raise idk.IsccUnsupportedMediatype(mediatype)

    cc.mediatype = mediatype
    cc.mode = mode
    cc.type_ = schema_org_map.get(mode)

    return cc


def code_text(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Text.

    Creates a Text-Code by extracting and processing text content from document files.
    Can optionally extract metadata and create a thumbnail representation of the text.

    :param fp: Filepath used for Text-Code creation.
    :key extract_meta: Whether to extract metadata. Default: True
    :key create_thumb: Whether to create a thumbnail. Default: True
    :key bits: Bit-length of the generated Text-Code UNIT. Default: 64
    :key granular: Whether to generate additional granular fingerprints. Default: False
    :return: ISCC metadata including Text-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.text_meta_extract(fp)

    if opts.create_thumb:
        thumbnail_img = idk.text_thumbnail(fp)
        if thumbnail_img:
            thumbnail_durl = idk.image_to_data_url(thumbnail_img)
            meta["thumbnail"] = thumbnail_durl

    text = idk.text_extract(fp)
    code = ic.gen_text_code_v0(text, bits=opts.bits)
    meta.update(code)
    if opts.granular:
        features = idk.text_features(ic.text_clean(text))
        meta["features"] = [features]
    return idk.IsccMeta.construct(**meta)


def code_image(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Image.

    Creates an Image-Code by normalizing and processing the visual content of image files.
    The image is normalized according to SDK options (transparency handling, border trimming, ...).

    :param fp: Filepath used for Image-Code creation.
    :key extract_meta: Whether to extract metadata. Default: True
    :key create_thumb: Whether to create a thumbnail. Default: True
    :key bits: Bit-length of the generated Image-Code UNIT. Default: 64
    :return: ISCC metadata including Image-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.image_meta_extract(fp)
    if opts.create_thumb:
        thumbnail_img = idk.image_thumbnail(fp)
        thumbnail_durl = idk.image_to_data_url(thumbnail_img)
        meta["thumbnail"] = thumbnail_durl

    pixels = idk.image_normalize(Image.open(fp))
    code_obj = ic.gen_image_code_v0(pixels, bits=opts.bits)
    meta.update(code_obj)

    return idk.IsccMeta.construct(**meta)


def code_audio(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Audio.

    Creates an Audio-Code by extracting acoustic fingerprints from audio files.
    Uses chromaprint/fpcalc to generate audio features for similarity matching.

    :param fp: Filepath used for Audio-Code creation.
    :key extract_meta: Whether to extract metadata. Default: True
    :key create_thumb: Whether to create a thumbnail. Default: True
    :key bits: Bit-length of the generated Audio-Code UNIT. Default: 64
    :return: ISCC metadata including Audio-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.audio_meta_extract(fp)
    if opts.create_thumb:
        thumbnail_img = idk.audio_thumbnail(fp)
        if thumbnail_img:
            thumbnail_durl = idk.image_to_data_url(thumbnail_img)
            meta["thumbnail"] = thumbnail_durl

    features = idk.audio_features_extract(fp)
    code_obj = ic.gen_audio_code_v0(features["fingerprint"], bits=opts.bits)
    meta.update(code_obj)

    return idk.IsccMeta.construct(**meta)


def code_video(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Generate Content-Code Video.

    Creates a Video-Code by extracting and processing visual features from video frames.
    Uses MPEG-7 signature tools to extract frame-based features and optionally detect scene changes.

    :param fp: Filepath used for Video-Code creation.
    :key extract_meta: Whether to extract metadata. Default: True
    :key create_thumb: Whether to create a thumbnail. Default: True
    :key granular: Generate additional fingerprints based on scenes. Default: False
    :key video_store_mp7sig: Whether to store extracted MP7 Video signature file. Default: False
    :key bits: Bit-length of the generated Video-Code UNIT. Default: 64
    :return: ISCC metadata including Video-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    meta = dict()

    if opts.extract_meta:
        meta = idk.video_meta_extract(fp)

    if opts.create_thumb:
        thumbnail_image = idk.video_thumbnail(fp)
        if thumbnail_image is not None:
            thumbnail_durl = idk.image_to_data_url(thumbnail_image)
            meta["thumbnail"] = thumbnail_durl

    sig, scenes = None, []
    if opts.granular:
        sig, scenes = idk.video_mp7sig_extract_scenes(fp)
    else:
        sig = idk.video_mp7sig_extract(fp)

    if opts.video_store_mp7sig:
        outp = fp.with_suffix(".iscc.mp7sig")
        with open(outp, "wb") as outf:
            outf.write(sig)

    frames = idk.read_mp7_signature(sig)
    features = [tuple(frame.vector.tolist()) for frame in frames]

    code_obj = ic.gen_video_code_v0(features, bits=opts.bits)
    meta.update(code_obj)

    if opts.granular:
        granular = idk.video_compute_granular(frames, scenes)
        meta["features"] = [granular]

    return idk.IsccMeta.construct(**meta)


def code_data(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Create ISCC Data-Code.

    The Data-Code is a similarity preserving hash of the raw input data that allows for
    detection of similar binary data regardless of file format or metadata differences.

    :param fp: Filepath used for Data-Code creation.
    :key bits: Bit-length of the generated Data-Code UNIT. Default: 64
    :return: ISCC metadata including Data-Code.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    with open(fp, "rb") as stream:
        meta = ic.gen_data_code_v0(stream, bits=opts.bits)

    return idk.IsccMeta.construct(**meta)


def code_instance(fp, **options):
    # type: (str|Path, Any) -> idk.IsccMeta
    """
    Create ISCC Instance-Code.

    The Instance-Code is a cryptographic hash (blake3) of the input data.
    Its purpose is to serve as a checksum that detects even minimal changes
    to the data of the referenced media asset. For cryptographically secure integrity
    checking, a full 256-bit multihash is provided with the `datahash` field.

    :param fp: Filepath used for Instance-Code creation.
    :key bits: Bit-length of the generated Instance-Code UNIT. Default: 64
    :return: ISCC metadata including Instance-Code, datahash and filesize.
    """
    fp = Path(fp)
    opts = idk.sdk_opts.override(options)
    with open(fp, "rb") as stream:
        meta = ic.gen_instance_code_v0(stream, bits=opts.bits)

    return idk.IsccMeta.construct(**meta)



"""IPFS wrapper"""

import sys
from pathlib import Path
import iscc_sdk as idk


__all__ = [
    "ipfs_cidv1",
    "ipfs_cidv1_base16",
]


def ipfs_cidv1(fp, wrap=False):
    # type: (str|Path) -> str
    """
    Create default IPFS CIDv1 for file at filepath `fp`.

    If `wrap` is True, the file will be wrapped with a directory and the filname will be appended
    to the directory CIDv1 with a `/`.

    :param fp: Filepath
    :return: IPFS CIDv1 of the file
    """
    fp = Path(fp)
    args = ["add", "--only-hash", "--cid-version=1", "--offline", "--quieter"]
    if wrap:
        args.append("--wrap-with-directory")
    args.append(fp)
    result = idk.run_ipfs(args)
    encoding = sys.stdout.encoding or "utf-8"
    cid = result.stdout.decode(encoding).strip()
    if wrap:
        cid += f"/{fp.name}"
    return cid


def ipfs_cidv1_base16(fp):
    # type: (str|Path) -> str
    """
    Create IPFS CIDv1 with base16 encoding.

    :param fp: Filepath
    :return: IPFS CIDv1 of the file in base16 (hex)
    """
    fp = Path(fp)
    args = [
        "add",
        "--only-hash",
        "--cid-version=1",
        "--offline",
        "--quieter",
        "--cid-base=base16",
        fp,
    ]
    result = idk.run_ipfs(args)
    encoding = sys.stdout.encoding or "utf-8"
    return result.stdout.decode(encoding).strip()



"""*Generate thumbnails for media assets*"""

from pathlib import Path
from PIL import Image
import iscc_sdk as idk

__all__ = [
    "thumbnail",
]

THUMBNAILERS = {
    "image": idk.image_thumbnail,
    "video": idk.video_thumbnail,
    "text": idk.text_thumbnail,
    "audio": idk.audio_thumbnail,
}


def thumbnail(fp):
    # type: (str|Path) -> Image.Image|None
    """
    Create a thumbnail for a media asset.

    :param fp: Filepath to media file.
    :return: Thumbnail image as PIL Image object
    """
    fp = Path(fp)
    mime, mode = idk.mediatype_and_mode(fp)
    thumbnailer = THUMBNAILERS.get(mode)
    if thumbnailer:
        return thumbnailer(fp)



"""*Detect and map RFC6838 mediatypes to ISCC processing modes*."""

from pathlib import Path
from loguru import logger as log
from typing import List, Optional, Union
import mimetypes
import magic
import iscc_sdk as idk


__all__ = [
    "mediatype_and_mode",
    "mediatype_guess",
    "mediatype_normalize",
    "mediatype_supported",
    "mediatype_clean",
    "mediatype_to_mode",
    "mediatype_from_name",
    "mediatype_from_data",
    "SUPPORTED_MEDIATYPES",
    "SUPPORTED_EXTENSIONS",
]


def mediatype_and_mode(fp):
    # type: (str|Path) -> tuple[str, str]
    """
    Detect mediatype and processing mode for a file.

    !!! example
        ```
        >>> import iscc_sdk
        >>> iscc_sdk.mediatype_and_mode('some.pdf')
        ('application/pdf', 'text')

        ```

    :param fp: Filepath
    :return: A tuple of `mediatype` and `mode`
    """
    fp = Path(fp)
    with open(fp, "rb") as infile:
        data = infile.read(4096)

    mediatype = mediatype_guess(data, file_name=fp.name)
    try:
        mode = mediatype_to_mode(mediatype)
    except idk.IsccUnsupportedMediatype:
        raise idk.IsccUnsupportedMediatype(f"Unsupported mediatype {mediatype} for {fp.name}")
    return mediatype, mode


def mediatype_guess(data, file_name=None):
    # type: (bytes, Optional[str]) -> str
    """
    Guess mediatype from raw data or filename.

    First try to guess by file extension. If that fails we match by content sniffing.

    !!! example
        ```
        >>> import iscc_sdk
        >>> iscc_sdk.mediatype_guess(b'GIF89a')
        'image/gif'

        ```

    :param bytes data: Raw file data (first 4096 bytes recommended)
    :param Optional[str] file_name: Filename for guessing based on file extension
    :return: Media type sting
    :rtype: str
    """
    guess_name, guess_data = None, None

    if file_name:
        guess_name = mediatype_from_name(file_name)

    guess_data = mediatype_from_data(data)

    # Normalize
    guess_data = mediatype_normalize(guess_data)
    guess_name = mediatype_normalize(guess_name)
    media_type = guess_name or guess_data

    # Special cases of missdetection
    if guess_data and "ogg" in guess_data:
        media_type = guess_data

    if media_type is None:
        log.warning("Could not detect mediatype. Fallback to application/octet-stream")
        return "application/octet-stream"  # Generic binary data as fallback

    log.debug(f"{media_type} mediatype detected")

    return media_type


def mediatype_normalize(mime):
    # type: (str) -> str
    """
    Normalize mediatype string.

    !!! example
        ```
        >>> import iscc_sdk
        >>> iscc_sdk.mediatype_normalize("audio/x-aiff")
        '"audio/aiff"'

        ```

    :param str mime: Mediatype sting
    :return: Normalized mediatype string
    :rtype: str
    """
    return MEDIATYPE_NORM.get(mime, mime)


def mediatype_supported(mime):
    # type: (str) -> bool
    """
    Check if mediatype is supported.

    :param str mime: Mediatype sting
    :return: True if mediatype is supported
    :rtype: bool
    """
    return mediatype_normalize(mime) in SUPPORTED_MEDIATYPES


def mediatype_from_name(name):
    # type: (str) -> Optional[str]
    """
    Guess mediatype from filename or URL.

    :param str name: Filename or URL
    :return: Mediatype string
    :rtype: str
    """
    return mimetypes.guess_type(name)[0]


def mediatype_from_data(data):
    # type: (bytes) -> Optional[str]
    """
    Guess mediatype by sniffing raw header data.

    :param bytes data: Raw fileheader data (first 4096 bytes recommended)
    :return: Mediatype string
    :rtype: str
    """
    try:
        return magic.from_buffer(data, mime=True)
    except Exception as e:
        log.warning((f"Failed mediatype sniffing: {e}"))
        return None


def mediatype_clean(mime):
    # type: (Union[str, List]) -> str
    """
    Clean mediatype/content-type string or first entry of a list of mimetype strings.

    Also removes semicolon separated encoding information.

    :param str|List mime: Mediatype or list of mediatypes
    :return: Mediatype string
    :rtype: str
    """
    if mime and isinstance(mime, List):
        mime = mime[0]
    if mime:
        mime = mime.split(";")[0]
    return mime.strip()


def mediatype_to_mode(mime_type):
    # type: (str) -> str
    """Get perceptual processing mode from mimetype.

    :param mime_type: RFC-6838 mediatype string
    :return Processing mode ("text", "image", "audio", "video")
    :raise IsccUnsupportedMediatype: if no matching processing mode was found.
    """

    mime_type = mediatype_clean(mime_type)
    entry = SUPPORTED_MEDIATYPES.get(mime_type)
    if entry:
        return entry["mode"]

    # Fallback to guess mode by top-level type
    mode = mime_type.split("/")[0]
    if mode in ["text", "image", "audio", "video"] and mime_type not in UNSUPPORTED_MEDIATYPES:
        log.warning(f"Guessing perceptual mode from {mime_type}")
        return mode

    raise idk.IsccUnsupportedMediatype(f"No known processing mode for {mime_type}")


mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/x-mobipocket-ebook", ".mobi")
mimetypes.add_type("application/x-sqlite3", ".sqlite")
mimetypes.add_type("video/mp4", ".f4v")
mimetypes.add_type("video/h264", ".h264")
mimetypes.add_type("audio/x-wavpack", ".wv")
mimetypes.add_type(
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document", ".docx"
)
mimetypes.add_type("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", ".xlsx")
mimetypes.add_type("image/heic", ".heic")
mimetypes.add_type("image/avif", ".avif")


SUPPORTED_MEDIATYPES = {
    # Text Formats
    "application/rtf": {"mode": "text", "ext": "rtf"},
    "application/msword": {"mode": "text", "ext": "doc"},
    "application/pdf": {"mode": "text", "ext": "pdf"},
    "application/epub+zip": {"mode": "text", "ext": "epub"},
    "text/xml": {"mode": "text", "ext": "xml"},
    "application/json": {"mode": "text", "ext": "json"},
    "application/xhtml+xml": {"mode": "text", "ext": "xhtml"},
    "application/vnd.oasis.opendocument.text": {"mode": "text", "ext": "odt"},
    "text/html": {"mode": "text", "ext": "html"},
    "text/plain": {"mode": "text", "ext": "txt"},
    "application/x-ibooks+zip": {"mode": "text", "ext": "ibooks"},
    "text/markdown": {"mode": "text", "ext": ["md", "markdown"]},
    "application/vnd.openxmlformats-officedocument.wordprocessingml.document": {
        "mode": "text",
        "ext": "docx",
    },
    "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet": {
        "mode": "text",
        "ext": "xlsx",
    },
    # Note: pptx only detected by file extension. Sniffing gives 'application/zip'
    "application/vnd.openxmlformats-officedocument.presentationml.presentation": {
        "mode": "text",
        "ext": "pptx",
    },
    "application/vnd.ms-powerpoint": {
        "mode": "text",
        "ext": "ppt",
    },
    "application/vnd.ms-excel": {"mode": "text", "ext": "xls"},
    # "application/x-mobipocket-ebook": {
    #     "mode": "text",
    #     "ext": ["mobi", "prc", "azw", "azw3", "azw4"],
    # },
    # Image Formats
    "image/bmp": {"mode": "image", "ext": "bmp"},
    "image/gif": {"mode": "image", "ext": "gif"},
    "image/jpeg": {"mode": "image", "ext": ["jpg", "jpeg"]},
    "image/png": {"mode": "image", "ext": "png"},
    "image/tiff": {"mode": "image", "ext": "tif"},
    "image/vnd.adobe.photoshop": {"mode": "image", "ext": "psd"},
    "application/postscript": {"mode": "image", "ext": "eps"},
    "image/avif": {"mode": "image", "ext": "avif"},
    "image/heic": {"mode": "image", "ext": "heic"},
    # Audio Formats
    "audio/mpeg": {"mode": "audio", "ext": "mp3"},
    "audio/wav": {"mode": "audio", "ext": "wav"},
    "audio/x-wav": {"mode": "audio", "ext": "wav"},
    "audio/ogg": {"mode": "audio", "ext": "ogg"},
    "audio/aiff": {"mode": "audio", "ext": "aif"},
    "audio/x-aiff": {"mode": "audio", "ext": "aif"},
    "audio/x-flac": {"mode": "audio", "ext": "flac"},
    "audio/opus": {"mode": "audio", "ext": "opus"},
    "audio/x-wavpack": {"mode": "audio", "ext": "wv"},
    # Video Formats
    "application/vnd.rn-realmedia": {"mode": "video", "ext": "rm"},
    "video/x-dirac": {"mode": "video", "ext": "drc"},
    "video/3gpp": {"mode": "video", "ext": "3gp"},
    "video/3gpp2": {"mode": "video", "ext": "3g2"},
    "video/x-ms-asf": {"mode": "video", "ext": "asf"},
    "video/avi": {"mode": "video", "ext": "avi"},
    "video/webm": {"mode": "video", "ext": "webm"},
    "video/mpeg": {"mode": "video", "ext": ["mpeg", "mpg", "m1v", "vob"]},
    "video/mp4": {"mode": "video", "ext": "mp4"},
    "video/x-m4v": {"mode": "video", "ext": "m4v"},
    "video/x-matroska": {"mode": "video", "ext": "mkv"},
    "video/ogg": {"mode": "video", "ext": ["ogg", "ogv"]},
    "video/quicktime": {"mode": "video", "ext": ["mov", "f4v"]},
    "video/x-flv": {"mode": "video", "ext": "flv"},
    "application/x-shockwave-flash": {"mode": "video", "ext": "swf"},
    "video/h264": {"mode": "video", "ext": "h264"},
    "video/x-ms-wmv": {"mode": "video", "ext": "wmv"},
}

# Signals eplixitly unsupported mediatypes that fail processing
UNSUPPORTED_MEDIATYPES = {
    "image/svg+xml": {"mode": "image", "ext": "svg"},
}


MEDIATYPE_NORM = {
    "audio/x-aiff": "audio/aiff",
    "audio/x-wav": "audio/wav",
    "image/x-ms-bmp": "image/bmp",
    "video/x-msvideo": "video/avi",
    "application/ogg": "video/ogg",
    "application/xml": "text/xml",
    "application/vnd.ms-asf": "video/x-ms-asf",
    "application/vnd.adobe.flash.movie": "application/x-shockwave-flash",
}

SUPPORTED_EXTENSIONS = []
for v in SUPPORTED_MEDIATYPES.values():
    ext = v["ext"]
    if isinstance(ext, str):
        SUPPORTED_EXTENSIONS.append(ext)
    else:
        for e in ext:
            SUPPORTED_EXTENSIONS.append(e)



"""SDK options can be configured using environment variables. Variables are defined as
class-atributes on the `SdkOptions` instance.

!!! example "Example how to access configuration options"
    ```python
    import iscc_sdk as idk

    # To access ISCC_SDK_VIDEO_FPS setting use
    fps: int = idk.sdk_opts.video_fps
    ```


"""

from typing import Optional, Literal
import copy

try:
    from pydantic.v1 import Field, validator, BaseSettings
except ImportError:  # pragma: no cover
    from pydantic import Field, validator, BaseSettings
from PIL import Image
import iscc_core


__all__ = [
    "SdkOptions",
    "sdk_opts",
    "core_opts",
]


class SdkOptions(BaseSettings):
    """SDK Configuration Options"""

    class Config:
        validate_assignment = True
        env_prefix = "ISCC_SDK_"
        env_file = "iscc-sdk.env"
        env_file_encoding = "utf-8"

    granular: bool = Field(
        default=False,
        description="ISCC_SDK_GRANULAR - Generate additional granular fingerprints for ISCC-CODES",
    )

    process_container: bool = Field(
        default=False,
        description="ISCC_SDK_PROCESS_CONTAINER - Process embedded elements in container formats",
    )

    min_image_size: int = Field(
        default=256,
        description="ISCC_SDK_MIN_IMAGE_SIZE - Minimum image size in pixels for container processing",
    )

    add_units: bool = Field(
        default=False,
        description="ISCC_SDK_ADD_UNITS - Add ISCC-UNITS to ISCC metadata (in 'units' property)",
    )

    bits: int = Field(
        default=64,
        description="ISCC_SDK_BITS - Bit size of ISCC-UNITS in ISCC metadata (in 'units' property)",
    )

    wide: bool = Field(
        default=False,
        description="ISCC_SDK_WIDE - For ISCC-SUM with Data & Instance codes only support wide mode",
    )

    extract_meta: bool = Field(
        default=True,
        description="ISCC_EXTRACT_META - Extract metadata from digital assets (defaut: True)",
    )

    create_meta: bool = Field(
        default=True,
        description="ISCC_CREATE_META - Create Meta-Code from embedded metadata (defaut: True)",
    )

    create_thumb: bool = Field(
        default=True,
        description="ISCC_CREATE_THUMBNAIL - Create thumbail for digital assets (defaut: True)",
    )

    image_exif_transpose: bool = Field(
        default=True,
        description="ISCC_SDK_IMAGE_EXIF_TRANSPOSE - Transpose image according to EXIF Orientation tag",
    )

    image_fill_transparency: bool = Field(
        default=True,
        description="ISCC_SDK_IMAGE_FILL_TRANSPARENCY - Add white background to image if it has alpha transparency",
    )

    image_trim_border: bool = Field(
        default=True, description="ISCC_SDK_IMAGE_TRIM_BORDER - Crop empty borders of images"
    )

    image_thumbnail_size: int = Field(
        default=128,
        description="ISCC_SDK_IMAGE_THUMBNAIL_SIZE - Size of larger side of thumbnail in number of pixels",
    )

    image_thumbnail_quality: int = Field(
        default=30,
        description="ISCC_SDK_IMAGE_THUMBNAIL_QUALITY - Thumbnail image compression setting (0-100)",
    )

    image_thumbnail_format: Literal["JPEG", "WEBP", "AVIF"] = Field(
        default="WEBP",
        description="ISCC_SDK_IMAGE_THUMBNAIL_FORMAT - Format for thumbnail images (JPEG, WEBP, or AVIF)",
    )

    image_max_pixels: Optional[int] = Field(
        default=128000000,
        description="ISCC_SDK_IMAGE_MAX_PIXELS - Maximum number of pixels allowed for processing (default 128Mpx / 0.5GB RGB)",
    )

    text_avg_chunk_size: int = Field(
        default=512,
        description="ISCC_SDK_TEXT_AVG_CHUNK_SIZE - Avg number of characters per text chunk for granular fingerprints",
    )

    video_fps: int = Field(
        default=5,
        description="ISCC_SDK_VIDEO_FPS - Frames per second to process for video hash (ignored when 0).",
    )

    video_scene_limit: float = Field(
        default=0.4,
        description="ISCC_SDK_VIDEO_SCENE_LIMIT - Threshold value above which a scene cut is created (default 0.4)",
    )

    video_store_mp7sig: bool = Field(
        default=False,
        description="ISCC_SDK_VIDEO_STORE_MP7SIG - Store extracted MP7 Video as <videofile>.iscc.mp7sig",
    )

    fallback: bool = Field(
        default=False,
        description="ISCC_SDK_FALLBACK - Create 2-UNIT ISCC-SUM for unsupported media types",
    )

    @validator("image_max_pixels")
    def set_pillow(cls, v):
        Image.MAX_IMAGE_PIXELS = v
        return v

    def override(self, update=None):
        # type: (dict|None) -> SdkOptions
        """Returns an updated and validated deep copy of the current settings instance."""

        update = update or {}  # sets {} if update is None

        opts = copy.deepcopy(self)
        # We need update fields individually so validation gets triggered
        for field, value in update.items():
            if hasattr(self, field):
                setattr(opts, field, value)
            else:
                raise ValueError(f"Invalid field: {field}")
        return opts


sdk_opts = SdkOptions()
core_opts = iscc_core.core_opts



"""*Metadata handling functions*"""

from loguru import logger as log
from pathlib import Path
from typing import Optional, List, Dict, Any


try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cover
    from pydantic import validator

import iscc_sdk as idk
import iscc_core as ic
import iscc_schema as iss

__all__ = [
    "extract_metadata",
    "embed_metadata",
    "IsccMeta",
]

EXTRACTORS = {
    "text": idk.text_meta_extract,
    "image": idk.image_meta_extract,
    "audio": idk.audio_meta_extract,
    "video": idk.video_meta_extract,
}

EMBEDDERS = {
    "text": idk.text_meta_embed,
    "image": idk.image_meta_embed,
    "audio": idk.audio_meta_embed,
    "video": idk.video_meta_embed,
}


def extract_metadata(fp):
    # type: (str|Path) -> idk.IsccMeta
    """
    Extract metadata from file.

    :param fp: Filepath to media file.
    :return: Metadata mapped to IsccMeta schema
    """
    fp = Path(fp)
    mime, mode = idk.mediatype_and_mode(fp)
    extractor = EXTRACTORS.get(mode)
    if extractor:
        metadata = extractor(fp)
        return idk.IsccMeta.construct(**metadata)


def embed_metadata(fp, meta):
    # type: (str, iss.IsccMeta) -> Optional[str]
    """
    Embed metadata into a copy of the media file and return path to updated file.

    :param str fp: Filepath to source media file
    :param IsccMeta meta: Metadata to embed into media file
    :return: Filepath to the new media file with embedded metadata (None if no embedding supported)
    :rtype: str|None
    """
    mime, mode = idk.mediatype_and_mode(fp)
    embedder = EMBEDDERS.get(mode)
    if embedder:
        new_file_path = embedder(fp, meta)
        return new_file_path


class IsccMeta(iss.IsccMeta):
    """Custom IsccMeta with text trimming and recursive `parts` support"""

    parts: Optional[List[Dict[str, Any]]] = None

    @validator("name", pre=True)
    def trim_name(cls, v):
        return ic.text_trim(v, idk.core_opts.meta_trim_name)

    @validator("description", pre=True)
    def trim_description(cls, v):
        return ic.text_trim(v, idk.core_opts.meta_trim_description)



import os
from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Iterator, Optional, Tuple
from loguru import logger as log
import typer
from pathlib import Path
import iscc_core as ic
import iscc_sdk as idk
from rich.console import Console
from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    TransferSpeedColumn,
    TimeRemainingColumn,
    DownloadColumn,
)

console = Console()
app = typer.Typer(add_completion=False, no_args_is_help=True)


def _log_formatter(record: dict) -> str:  # pragma: no cover
    """Log message formatter"""
    color_map = {
        "TRACE": "blue",
        "DEBUG": "cyan",
        "INFO": "bold",
        "SUCCESS": "bold green",
        "WARNING": "yellow",
        "ERROR": "bold red",
        "CRITICAL": "bold white on red",
    }
    lvl_color = color_map.get(record["level"].name, "cyan")
    return (
        "[not bold green]{time:YYYY/MM/DD HH:mm:ss}[/not bold green] | {module:<12} | {line:<3} | {level.icon}"
        + f"  - [{lvl_color}]{{message}}[/{lvl_color}]"
    )


log.add(console.print, level="DEBUG", format=_log_formatter, colorize=True)


def iter_unprocessed(path, root_path=None):
    # type: (str|Path, Optional[str|Path]) -> Iterator[Tuple[Path, int]]
    """
    Walk directory tree recursively with deterministic ordering and yield tuples of file metadata.

    Metadata = (relpath, size)

    - path: pathlib.Path object
    - size: integer file size in number of bytes

    File-entries are yielded in reproducible and deterministic order (bottom-up). Symlink and
    processed files are ignored silently.

    Implementation Note: We use os.scandir to reduce the number of syscalls for metadata collection.
    """
    root_path = Path(root_path or path)
    with os.scandir(path) as entries:
        # Sort the entries
        sorted_entries = sorted(entries, key=lambda e: e.name)

        # Separate directories and files
        dirs = [entry for entry in sorted_entries if entry.is_dir()]
        files = [entry for entry in sorted_entries if entry.is_file()]

        # Recursively process directories first (bottom-up traversal)
        for dir_entry in dirs:
            yield from iter_unprocessed(Path(dir_entry.path), root_path=root_path)

        # Process files in the current directory
        for file_entry in files:
            file_path = Path(file_entry)
            # Ignore result files
            if file_path.name.endswith(".iscc.json") or file_path.name.endswith(".iscc.mp7sig"):
                continue
            # Ignore files that have results
            if Path(file_path.as_posix() + ".iscc.json").exists():
                continue
            file_size = file_entry.stat().st_size
            yield file_path, file_size


def process_file(fp: Path):
    idk.sdk_opts.video_store_mp7sig = True
    try:
        return fp, idk.code_iscc(fp.as_posix())
    except Exception as e:
        return fp, e


@app.command()
def create(file: Path):
    """Create ISCC-CODE for single FILE."""
    log.remove()
    if file.is_file() and file.exists():
        try:
            result = idk.code_iscc(file.as_posix())
            typer.echo(result.json(indent=2))
        except idk.IsccUnsupportedMediatype as e:
            typer.echo(e)
            raise typer.Exit(code=1)
    else:
        typer.echo(f"Invalid file path {file}")
        raise typer.Exit(code=1)


@app.command()
def batch(folder: Path, workers: int = os.cpu_count()):  # pragma: no cover
    """Create ISCC-CODEs for files in FOLDER (parallel & recursive)."""
    if not folder.is_dir() or not folder.exists():
        typer.echo(f"Invalid folder {folder}")
        raise typer.Exit(1)

    file_paths = []
    file_sizes = []
    for path, size in iter_unprocessed(folder):
        file_paths.append(path)
        file_sizes.append(size)

    file_sizes_dict = {path: size for path, size in zip(file_paths, file_sizes)}
    total_size = sum(file_sizes)
    progress = Progress(
        TextColumn("[bold blue]Processing {task.fields[dirname]}", justify="right"),
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
        console=console,
    )

    with progress:
        task_id = progress.add_task("Processing", dirname=folder.name, total=total_size)
        with ProcessPoolExecutor(max_workers=workers) as executor:
            futures = [executor.submit(process_file, fp) for fp in file_paths]
            for future in as_completed(futures):
                fp, iscc_meta = future.result()
                if isinstance(iscc_meta, idk.IsccMeta):
                    out_path = Path(fp.as_posix() + ".iscc.json")
                    with out_path.open(mode="wt", encoding="utf-8") as outf:
                        outf.write(iscc_meta.json(indent=2))
                    log.info(f"Finished {fp.name}")
                else:
                    log.error(f"Failed {fp.name}: {iscc_meta}")
                progress.update(task_id, advance=file_sizes_dict[fp], refresh=True)


@app.command()
def install():
    """Install content processing tools."""
    idk.install()


@app.command()
def selftest():
    """Run conformance tests."""
    ic.conformance_selftest()


@app.command()
def extract(file: Path):
    """Extract text from FILE and print to console."""
    log.remove()
    if file.is_file() and file.exists():
        try:
            text = idk.text_extract(file.as_posix())
            typer.echo(ic.text_clean(text))
        except Exception as e:
            typer.echo(f"Error extracting text: {e}")
            raise typer.Exit(code=1)
    else:
        typer.echo(f"Invalid file path {file}")
        raise typer.Exit(code=1)


if __name__ == "__main__":  # pragma: no cover
    app()



"""ISCC - Software Development Kit."""

import os
from importlib import metadata
from platformdirs import PlatformDirs


__version__ = metadata.version("iscc-sdk")
APP_NAME = "iscc-sdk"
APP_AUTHOR = "iscc"
dirs = PlatformDirs(appname=APP_NAME, appauthor=APP_AUTHOR)
os.makedirs(dirs.user_data_dir, exist_ok=True)
os.environ["LOGURU_AUTOINIT"] = "False"


# Import full api to toplevel
from iscc_sdk.monkeys import *
from iscc_sdk.compat import *
from iscc_sdk.options import *
from iscc_sdk.tools import *
from iscc_sdk.mediatype import *
from iscc_sdk.container import *
from iscc_sdk.image import *
from iscc_sdk.main import *
from iscc_sdk.text import *
from iscc_sdk.ipfs import *
from iscc_sdk.audio import *
from iscc_sdk.video import *
from iscc_sdk.mp7 import *
from iscc_sdk.exceptions import *
from iscc_sdk.metadata import *
from iscc_sdk.pdf import *
from iscc_sdk.epub import *
from iscc_sdk.docx_ import *
from iscc_sdk.thumbnail import *
from iscc_sdk.utils import *



"""*Image handling module*."""

from pathlib import Path
import exiv2
import base64
import io
import shutil
import tempfile
from typing import Sequence
from iscc_schema import IsccMeta
from loguru import logger as log
from PIL import Image, ImageEnhance, ImageChops, ImageOps
import iscc_sdk as idk
from pillow_heif import register_heif_opener

register_heif_opener()


__all__ = [
    "image_normalize",
    "image_exif_transpose",
    "image_fill_transparency",
    "image_trim_border",
    "image_meta_embed",
    "image_meta_extract",
    "image_meta_delete",
    "image_strip_metadata",
    "image_thumbnail",
    "image_to_data_url",
]


Image.MAX_IMAGE_PIXELS = idk.sdk_opts.image_max_pixels


def image_normalize(img):
    # type: (Image.Image) -> Sequence[int]
    """
    Normalize image for hash calculation.

    :param img: Pillow Image Object
    :return: Normalized and flattened image as 1024-pixel array (from 32x32 gray pixels)
    """

    # Transpose image according to EXIF Orientation tag
    if idk.sdk_opts.image_exif_transpose:
        img = image_exif_transpose(img)

    # Add white background to image if it has alpha transparency
    if idk.sdk_opts.image_fill_transparency:
        img = image_fill_transparency(img)

    # Trim uniform colored (empty) border if there is one
    if idk.sdk_opts.image_trim_border:
        img = image_trim_border(img)

    # Convert to grayscale
    img = img.convert("L")

    # Resize to 32x32
    im = img.resize((32, 32), idk.BICUBIC)

    # A flattened sequence of grayscale pixel values (1024 pixels)
    pixels = im.getdata()

    return pixels


def image_exif_transpose(img):
    # type: (Image.Image) -> Image.Image
    """
    Transpose image according to EXIF Orientation tag

    :param img: Pillow Image Object
    :return: EXIF transposed image
    """
    img = ImageOps.exif_transpose(img)
    log.debug(f"Image exif transpose applied")
    return img


def image_fill_transparency(img):
    # type: (Image.Image) -> Image.Image
    """
    Add white background to image if it has alpha transparency.

    :param img: Pillow Image Object
    :return: Image with transparency replaced by white background
    """
    if img.mode != "RGBA":
        img = img.convert("RGBA")
    bg = Image.new("RGBA", img.size, (255, 255, 255))
    img = Image.alpha_composite(bg, img)
    log.debug(f"Image transparency filled with white background")
    return img


def image_trim_border(img):
    # type: (Image.Image) -> Image.Image
    """Trim uniform colored (empty) border.

    Takes the upper left pixel as reference for border color.

    :param img: Pillow Image Object
    :return: Image with uniform colored (empty) border removed.
    """

    bg = Image.new(img.mode, img.size, img.getpixel((0, 0)))
    diff = ImageChops.difference(img, bg)
    diff = ImageChops.add(diff, diff)
    bbox = diff.getbbox()
    if bbox != (0, 0) + img.size:
        log.debug(f"Image has been trimmed")
        return img.crop(bbox)
    return img


def image_meta_extract(fp):
    # type: (str|Path) -> dict
    """
    Extract metadata from image using native exiv2 bindings.

    :param fp: Filepath to image file.
    :return: Metadata mapped to IsccMeta schema
    """
    fp = Path(fp)
    img_exiv = exiv2.ImageFactory.open(fp.as_posix())
    img_exiv.readMetadata()

    # Read and process all metadata types: EXIF, XMP, IPTC
    meta_dict = {}
    meta_dict.update(_process_metadata(img_exiv.exifData()))
    meta_dict.update(_process_metadata(img_exiv.xmpData(), is_xmp=True))
    meta_dict.update(_process_metadata(img_exiv.iptcData()))

    # Map metadata to schema fields
    mapped = {}
    for tag, mapped_field in IMAGE_META_MAP.items():
        if mapped_field in mapped:
            continue
        if tag in meta_dict and meta_dict[tag]:
            try:
                mapped[mapped_field] = idk.text_sanitize(meta_dict[tag])
            except Exception as e:  # pragma: no cover
                log.error(f"Failed to sanitize {meta_dict[tag]}: {e}")
                continue

    # Add image dimensions
    with Image.open(fp) as img:
        mapped["width"], mapped["height"] = img.size

    return mapped


def image_meta_embed(fp, meta):
    # type: (str|Path, IsccMeta) -> Path
    """
    Embed metadata into a copy of the image file.

    :param fp: Filepath to source image file
    :param meta: Metadata to embed into image
    :return: Filepath to the new image file with updated metadata
    """
    fp = Path(fp)

    # Create temp directory and copy the image
    tempdir = Path(tempfile.mkdtemp())
    imagefile = Path(shutil.copy(fp, tempdir))

    # Register XMP namespaces before opening image
    exiv2.XmpProperties.registerNs("http://purl.org/iscc/schema/", "iscc")
    exiv2.XmpProperties.registerNs("http://purl.org/dc/elements/1.1/", "dc")
    exiv2.XmpProperties.registerNs("http://ns.useplus.org/", "plus")
    exiv2.XmpProperties.registerNs("http://ns.adobe.com/xap/1.0/rights/", "xmpRights")

    # Open the copied image with exiv2
    img_exiv = exiv2.ImageFactory.open(str(imagefile))
    img_exiv.readMetadata()

    # Get metadata collections
    xmp_data = img_exiv.xmpData()

    # Set simple metadata values
    if meta.name:
        xmp_data["Xmp.iscc.name"] = meta.name
        xmp_data["Xmp.dc.title"] = meta.name
    if meta.description:
        xmp_data["Xmp.iscc.description"] = meta.description
        xmp_data["Xmp.dc.description"] = meta.description
    if meta.meta:
        xmp_data["Xmp.iscc.meta"] = meta.meta
    if meta.license:
        xmp_data["Xmp.xmpRights.WebStatement"] = meta.license
    if meta.creator:
        xmp_data["Xmp.dc.creator"] = meta.creator
    if meta.rights:
        xmp_data["Xmp.dc.rights"] = meta.rights
    if meta.identifier:
        xmp_data["Xmp.dc.identifier"] = meta.identifier

    # Set complex metadata values
    if meta.acquire:
        # Set the Licensor URL
        # First create a bag value
        licensor_bag = exiv2.XmpTextValue()
        licensor_bag.setXmpArrayType(exiv2.XmpValue.XmpArrayType.xaBag)
        xmp_data["Xmp.plus.Licensor"] = licensor_bag

        # Then set the LicensorURL with the struct path
        xmp_data["Xmp.plus.Licensor[1]/plus:LicensorURL"] = meta.acquire

    # Write metadata back to the file
    img_exiv.writeMetadata()

    log.debug(f"Embedding {meta.dict(exclude_unset=True)} in {imagefile.name}")
    return imagefile


def image_meta_delete(fp):
    # type: (str|Path) -> None
    """
    Delete all metadata from image.

    :param fp: Filepath to image file.
    """
    fp = Path(fp)
    img_exiv = exiv2.ImageFactory.open(str(fp))
    img_exiv.readMetadata()

    # Clear all metadata
    img_exiv.exifData().clear()
    img_exiv.xmpData().clear()
    img_exiv.iptcData().clear()

    # Write the cleared metadata back to the file
    img_exiv.writeMetadata()

    log.debug(f"Deleted all metadata from {fp.name}")


def image_thumbnail(fp):
    # type: (str|Path) -> Image.Image
    """
    Create a thumbnail for an image.

    :param fp: Filepath to image file.
    :return: Thumbnail image as PIL Image object
    """
    fp = Path(fp)
    size = idk.sdk_opts.image_thumbnail_size
    img = Image.open(fp)

    # Convert to RGB before resizing
    img = img.convert("RGB")

    img.thumbnail((size, size), resample=idk.LANCZOS)
    return ImageEnhance.Sharpness(img).enhance(1.4)


def image_strip_metadata(img):
    # type: (Image.Image) -> Image.Image
    """
    Strip all metadata from Pillow Image object.

    :param img: PIL Image object to strip metadata from.
    :return: Image.Image
    """
    data = list(img.getdata())
    new_img = Image.new(img.mode, img.size)
    new_img.putdata(data)
    return new_img


def image_to_data_url(img):
    # type: (Image.Image) -> str
    """
    Convert PIL Image object to WebP Data-URL.

    :param img: PIL Image object to encode as WebP Data-URL.
    :return: Data-URL string
    """
    format_ = idk.sdk_opts.image_thumbnail_format
    quality = idk.sdk_opts.image_thumbnail_quality
    img = image_strip_metadata(img)
    raw = io.BytesIO()
    img.save(raw, format=format_, quality=quality)
    raw.seek(0)
    enc = base64.b64encode(raw.read()).decode("ascii")
    return f"data:image/{format_.lower()};base64," + enc


def _clean_xmp_value(value):
    # type: (str) -> str
    """
    Clean XMP value by removing language qualifiers.

    :param value: XMP value string that might contain language qualifier
    :return: Cleaned value string
    """
    # Remove language qualifier like 'lang="x-default" '
    if value.startswith('lang="') and '"' in value[6:]:
        lang_end = value.find('"', 6)
        if lang_end > 0 and len(value) > lang_end + 2:
            return value[lang_end + 2 :].strip()
    return value


def _process_metadata(metadata, is_xmp=False):
    # type: (object, bool) -> dict
    """
    Process metadata from exiv2 data collections.

    :param metadata: exiv2 metadata collection (exifData, xmpData, or iptcData)
    :param is_xmp: Whether the metadata is XMP (for special processing)
    :return: Dictionary of processed metadata
    """
    result = {}
    for datum in metadata:
        key = datum.key()
        raw_value = datum.value
        value = raw_value() if callable(raw_value) else raw_value

        if not isinstance(value, str):
            if hasattr(value, "to_string"):
                value = value.to_string()
            else:
                value = str(value)

        value = value.strip()

        # Clean XMP values from language qualifiers
        if is_xmp:
            value = _clean_xmp_value(value)

        result[key] = value

    return result


IMAGE_META_MAP = {
    "Xmp.iscc.name": "name",
    "Xmp.iscc.description": "description",
    "Xmp.iscc.meta": "meta",
    "Xmp.dc.title": "name",
    "Xmp.xmp.Nickname": "name",
    "Xmp.xmpDM.shotName": "name",
    "Xmp.photoshop.Headline": "name",
    "Xmp.iptcExt.AOTitle": "name",
    "Iptc.Application2.Headline": "name",
    "Iptc.Application2.BylineTitle": "name",
    "Iptc.Application2.ObjectName": "name",
    "Exif.Image.XPTitle": "name",
    "Xmp.dc.description": "description",
    "Xmp.dc.creator": "creator",
    "Iptc.Application2.Byline": "creator",
    "Exif.Image.Artist": "creator",
    "Xmp.xmpRights.WebStatement": "license",
    "Xmp.plus.Licensor[0].plus.LicensorURL": "acquire",
    "Xmp.plus.Licensor[1]/plus:LicensorURL": "acquire",
    "Xmp.dc.rights": "rights",
    "Xmp.dc.identifier": "identifier",
    "Xmp.xmp.Identifier": "identifier",
    "Xmp.dc.language": "language",
    "Iptc.Application2.Language": "language",
    "Exif.Image.ImageID": "identifier",
    "Exif.Photo.ImageUniqueID": "identifier",
}



"""*Manage SDK binary media file handling tools*."""

import os
import platform
import shutil
import subprocess
import tarfile
import zipfile
from pathlib import Path
from platform import system, architecture
from typing import List
from urllib.parse import urlparse
from urllib.request import urlretrieve
from blake3 import blake3
from loguru import logger as log
import stat
import iscc_sdk as idk
from concurrent.futures import ThreadPoolExecutor


__all__ = [
    "install",
    "run_ffprobe",
    "run_ffmpeg",
    "run_fpcalc",
    "run_ipfs",
]

BASE_VERSION = "1.0.0"
BASE_URL = f"https://github.com/iscc/iscc-binaries/releases/download/v{BASE_VERSION}"

IPFS_VERSION = "0.12.0"
IPFS_URLS = {
    "windows-64": f"{BASE_URL}/go-ipfs_v{IPFS_VERSION}_windows-amd64.zip",
    "linux-64": f"{BASE_URL}/go-ipfs_v{IPFS_VERSION}_linux-amd64.tar.gz",
    "darwin-64": f"{BASE_URL}/go-ipfs_v{IPFS_VERSION}_darwin-amd64.tar.gz",
}
IPFS_CHECKSUMS = {
    "windows-64": "a2af645936a090c296b5d54755af0e9d6b1021f652195bb8fc596cf2073001c7",
    "linux-64": "7312b34bc7179c94c96e09a067b61d405672653bbaf70abee30396433a18ef81",
    "darwin-64": "3797fd0e6d5f922c095a12d860baccb49d90cef74accf49d219d4cea1b0d2bd7",
}

FFPROBE_VERSION = "6.1"
FFPROBE_URLS = {
    "windows-64": f"{BASE_URL}/ffprobe-{FFPROBE_VERSION}-win-64.zip",
    "linux-64": f"{BASE_URL}/ffprobe-{FFPROBE_VERSION}-linux-64.zip",
    "darwin-64": f"{BASE_URL}/ffprobe-{FFPROBE_VERSION}-macos-64.zip",
}
FFPROBE_CHECKSUMS = {
    "linux-64": "bd01b79f3a0d0a19cd0e086b38f60b4fad4d3274bb5cfebf8abc3f1da5603a16",
    "darwin-64": "799f36c0cd8797b015186e7aa78d1dd93a634f56e084325de2672d89a840eabb",
    "windows-64": "75497ab05dad5b8e7edb52ad1e1b626c8756fe33ae67825ab202b213f6d80548",
}

FFMPEG_VERSION = "6.1"
FFMPEG_URLS = {
    "windows-64": f"{BASE_URL}/ffmpeg-{FFMPEG_VERSION}-win-64.zip",
    "linux-64": f"{BASE_URL}/ffmpeg-{FFMPEG_VERSION}-linux-64.zip",
    "darwin-64": f"{BASE_URL}/ffmpeg-{FFMPEG_VERSION}-osx-64.zip",
}
FFMPEG_CHECKSUMS = {
    "linux-64": "e96b9796ad7404151eabaaaea30151fb9226554ee87d5164161fd20de4482dd1",
    "darwin-64": "ba9313f7bf8c46ebe60b8f39fa6de1657c94f51fd14d77c6e02ef09fab6ff5bc",
    "windows-64": "06814a07bff1f281a5282110ebff2ccb088cd9ffda35fce5af63b8e88cc2385a",
}

FPCALC_VERSION = "1.5.1"
FPCALC_URLS = {
    "windows-64": f"{BASE_URL}/chromaprint-fpcalc-{FPCALC_VERSION}-windows-x86_64.zip",
    "linux-64": f"{BASE_URL}/chromaprint-fpcalc-{FPCALC_VERSION}-linux-x86_64.tar.gz",
    "darwin-64": f"{BASE_URL}/chromaprint-fpcalc-{FPCALC_VERSION}-macos-x86_64.tar.gz",
}
FPCALC_CHECKSUMS = {
    "windows-64": "e29364a879ddf7bea403b0474a556e43f40d525e0d8d5adb81578f1fbf16d9ba",
    "linux-64": "190977d9419daed8a555240b9c6ddf6a12940c5ff470647095ee6242e217de5c",
    "darwin-64": "afea164b0bc9b91e5205d126f96a21836a91ea2d24200e1b7612a7304ea3b4f1",
}


def install():
    """Install binary tools for content extraction and metadata handling."""
    with ThreadPoolExecutor(max_workers=6) as p:
        p.submit(fpcalc_install)
        p.submit(ffprobe_install)
        p.submit(ffmpeg_install)
        p.submit(ipfs_install)
    return True


def system_tag():
    os_tag = system().lower()
    os_bits = architecture()[0].rstrip("bit")
    return f"{os_tag}-{os_bits}"


def is_installed(fp: str) -> bool:
    """Check if binary at `fp` exists and is executable."""
    return os.path.isfile(fp) and os.access(fp, os.X_OK)


def extract(archive):  # pragma: no cover
    """Extract downloded archive."""

    if archive.endswith(".zip"):
        with zipfile.ZipFile(archive, "r") as zip_file:
            zip_file.extractall(Path(archive).parent.absolute())

    elif archive.endswith("tar.gz"):
        with tarfile.open(archive, "r:gz") as tar_file:

            def is_within_directory(directory, target):
                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)

                prefix = os.path.commonprefix([abs_directory, abs_target])

                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(tar_file, Path(archive).parent.absolute())
    os.unlink(archive)


########################################################################################
# IPFS                                                                                 #
########################################################################################


def ipfs_download_url() -> str:
    """Return system and version dependent IPFS download url."""
    return IPFS_URLS[system_tag()]


def ipfs_bin() -> str:
    """Returns local path to IPFS executable."""
    path = os.path.join(idk.dirs.user_data_dir, "go-ipfs", "ipfs")
    if system() == "Windows":
        path += ".exe"
    return path


def ipfs_download():  # pragma: no cover
    """Download IPFS and return path to archive file."""
    b3 = IPFS_CHECKSUMS.get(system_tag())
    return download_file(ipfs_download_url(), checksum=b3)


def ipfs_install():  # pragma: no cover
    """Install IPFS cli tool and return path to executble"""
    if is_installed(ipfs_bin()):
        log.debug("ipfs is already installed")
        return ipfs_bin()
    log.critical("installing ipfs")
    archive_path = ipfs_download()
    extract(archive_path)
    st = os.stat(ipfs_bin())
    os.chmod(ipfs_bin(), st.st_mode | stat.S_IEXEC)
    # Initialize ipfs repo
    subprocess.run([ipfs_bin(), "init"])
    return ipfs_bin()


def ipfs_version_info():  # pragma: no cover
    """Get IPFS version"""
    try:
        r = subprocess.run([ipfs_bin(), "--version"], stdout=subprocess.PIPE)
        return r.stdout.decode("utf-8").splitlines()[0].strip()
    except FileNotFoundError:
        return "IPFS not installed"


def run_ipfs(args: List[str]):
    """Run ipfs command with `args`. Install ipfs if not found."""
    cmd = [ipfs_bin()] + args
    try:
        result = subprocess.run(cmd, capture_output=True, check=True)
    except FileNotFoundError:  # pragma: no cover
        print("IPFS not found - installing ...")
        ipfs_install()
        result = subprocess.run(cmd, capture_output=True, check=True)
    return result


########################################################################################
# Fpcalc                                                                               #
########################################################################################


def fpcalc_bin():  # pragma: no cover
    """Returns local path to fpcalc executable."""
    if system() == "Windows":
        return os.path.join(idk.dirs.user_data_dir, "fpcalc-{}.exe".format(FPCALC_VERSION))
    return os.path.join(idk.dirs.user_data_dir, "fpcalc-{}".format(FPCALC_VERSION))


def fpcalc_is_installed():  # pragma: no cover
    """Check if fpcalc is installed."""
    fp = fpcalc_bin()
    return os.path.isfile(fp) and os.access(fp, os.X_OK)


def fpcalc_download_url():
    """Return system and version dependent download url."""
    return FPCALC_URLS[system_tag()]


def fpcalc_download():  # pragma: no cover
    """Download fpcalc and return path to archive file."""
    b3 = FPCALC_CHECKSUMS.get(system_tag())
    return download_file(fpcalc_download_url(), checksum=b3)


def fpcalc_extract(archive):  # pragma: no cover
    """Extract archive with fpcalc executable."""
    if archive.endswith(".zip"):
        with zipfile.ZipFile(archive, "r") as zip_file:
            for member in zip_file.namelist():
                filename = os.path.basename(member)
                if filename == "fpcalc.exe":
                    source = zip_file.open(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)
    elif archive.endswith("tar.gz"):
        with tarfile.open(archive, "r:gz") as tar_file:
            for member in tar_file.getmembers():
                if member.isfile() and member.name.endswith("fpcalc"):
                    source = tar_file.extractfile(member)
                    target = open(fpcalc_bin(), "wb")
                    with source, target:
                        shutil.copyfileobj(source, target)
    os.unlink(archive)


def fpcalc_install():  # pragma: no cover
    """Install fpcalc command line tool and return path to executable."""
    if fpcalc_is_installed():
        log.debug("Fpcalc is already installed.")
        return fpcalc_bin()
    log.critical("installing fpcalc")
    archive_path = fpcalc_download()
    fpcalc_extract(archive_path)
    st = os.stat(fpcalc_bin())
    os.chmod(fpcalc_bin(), st.st_mode | stat.S_IEXEC)
    return fpcalc_bin()


def fpcalc_version_info():  # pragma: no cover
    """Get fpcalc version."""
    try:
        r = subprocess.run([fpcalc_bin(), "-v"], stdout=subprocess.PIPE)
        return r.stdout.decode("utf-8").strip().split()[2]
    except FileNotFoundError:
        return "FPCALC not installed"


def run_fpcalc(args: List[str]):
    """Run fpcalc command with `args`. Installs fpcalc if not found."""
    cmd = [fpcalc_bin()] + args
    try:
        result = subprocess.run(cmd, capture_output=True, check=True)
    except FileNotFoundError:  # pragma: no cover
        print("FPCALC not found - installing ...")
        fpcalc_install()
        result = subprocess.run(cmd, capture_output=True, check=True)
    return result


########################################################################################
# ffprobe                                                                              #
########################################################################################


def ffprobe_download_url():
    """Return system dependent download url."""
    return FFPROBE_URLS[system_tag()]


def ffprobe_bin() -> str:
    """Returns local path to ffprobe executable."""
    path = os.path.join(idk.dirs.user_data_dir, "ffprobe-{}".format(FFPROBE_VERSION))
    if system() == "Windows":
        path += ".exe"
    return path


def ffprobe_download():  # pragma: no cover
    """Download ffprobe and return path to archive file."""
    b3 = FFPROBE_CHECKSUMS.get(system_tag())
    return download_file(ffprobe_download_url(), checksum=b3)


def ffprobe_extract(archive: str):  # pragma: no cover
    """Extract ffprobe from archive."""
    fname = "ffprobe.exe" if system() == "Windows" else "ffprobe"
    with zipfile.ZipFile(archive) as zip_file:
        with zip_file.open(fname) as zf, open(ffprobe_bin(), "wb") as lf:
            shutil.copyfileobj(zf, lf)
    os.unlink(archive)


def ffprobe_install():  # pragma: no cover
    """Install ffprobe command line tool and return path to executable."""
    if is_installed(ffprobe_bin()):
        log.debug("ffprobe is already installed")
        return ffprobe_bin()
    log.critical("installing ffprobe")
    archive_path = ffprobe_download()
    ffprobe_extract(archive_path)
    st = os.stat(ffprobe_bin())
    os.chmod(ffprobe_bin(), st.st_mode | stat.S_IEXEC)
    assert is_installed(ffprobe_bin())
    return ffprobe_bin()


def ffprobe_version_info():  # pragma: no cover
    """Get ffprobe version"""
    try:
        r = subprocess.run([ffprobe_bin(), "-version"], stdout=subprocess.PIPE)
        return (
            r.stdout.decode("utf-8")
            .strip()
            .splitlines()[0]
            .split()[2]
            .rstrip("-static")
            .rstrip("-tessu")
        )
    except FileNotFoundError:
        return "ffprobe not installed"


def run_ffprobe(args: List[str]):  # pragma: no cover
    """Run ffprobe command with `args`. Install ffprobe if not found."""
    cmd = [ffprobe_bin()] + args
    try:
        result = subprocess.run(cmd, capture_output=True, check=True)
    except FileNotFoundError:  # pragma: no cover
        print("FFPROBE not found - installing ...")
        ffprobe_install()
        result = subprocess.run(cmd, capture_output=True, check=True)
    return result


########################################################################################
# ffmpeg                                                                               #
########################################################################################


def ffmpeg_download_url():
    """Return system dependent download url."""
    return FFMPEG_URLS[system_tag()]


def ffmpeg_bin() -> str:
    """Returns local path to ffmpeg executable."""
    path = os.path.join(idk.dirs.user_data_dir, "ffmpeg-{}".format(FFMPEG_VERSION))
    if system() == "Windows":
        path += ".exe"
    return path


def ffmpeg_download():  # pragma: no cover
    """Download ffmpeg and return path to archive file."""
    b3 = FFMPEG_CHECKSUMS.get(system_tag())
    return download_file(ffmpeg_download_url(), checksum=b3)


def ffmpeg_extract(archive: str):  # pragma: no cover
    """Extract ffmpeg from archive."""
    fname = "ffmpeg.exe" if system() == "Windows" else "ffmpeg"
    with zipfile.ZipFile(archive) as zip_file:
        with zip_file.open(fname) as zf, open(ffmpeg_bin(), "wb") as lf:
            shutil.copyfileobj(zf, lf)
    os.unlink(archive)


def ffmpeg_install():  # pragma: no cover
    """Install ffmpeg command line tool and return path to executable."""
    if is_installed(ffmpeg_bin()):
        log.debug("ffmpeg is already installed")
        return ffmpeg_bin()
    log.critical("installing ffmpeg")
    archive_path = ffmpeg_download()
    ffmpeg_extract(archive_path)
    st = os.stat(ffmpeg_bin())
    os.chmod(ffmpeg_bin(), st.st_mode | stat.S_IEXEC)
    return ffmpeg_bin()


def ffmpeg_version_info():  # pragma: no cover
    """Get ffmpeg version."""
    try:
        r = subprocess.run([ffmpeg_bin(), "-version"], stdout=subprocess.PIPE)
        return (
            r.stdout.decode("utf-8")
            .strip()
            .splitlines()[0]
            .split()[2]
            .rstrip("-static")
            .rstrip("-tessu")
        )
    except FileNotFoundError:
        return "ffmpeg not installed"


def run_ffmpeg(args: List[str]):
    """Run ffmpeg command with `args`. Install ffmpeg if not found."""
    cmd = [ffmpeg_bin()] + args
    try:
        result = subprocess.run(cmd, capture_output=True, check=True)
    except FileNotFoundError:  # pragma: no cover
        print("FFMPEG not found - installing ...")
        ffmpeg_install()
        result = subprocess.run(cmd, capture_output=True, check=True)
    return result


def download_file(url, checksum):  # pragma: no cover
    # type: (str, str) -> str
    """Download file to app directory and return path to downloaded file."""
    url_obj = urlparse(url)
    if not url_obj.scheme == "https":
        raise ValueError("Only https connections supported.")
    file_name = os.path.basename(url_obj.path)
    out_dir = idk.dirs.user_data_dir
    out_path = os.path.join(out_dir, file_name)
    if os.path.exists(out_path):
        log.debug(f"{file_name} already exists. Checking integrity")
        b3_calc = blake3(open(out_path, "rb").read()).hexdigest()
        if not checksum == b3_calc:
            log.critical(f"Integrity error for {out_path}. Redownloading")
        else:
            log.debug(f"{file_name} integrity ok - skipping redownload")
            return out_path
    log.debug(f"downloading {url} to {out_path}")
    urlretrieve(url, filename=out_path)
    log.debug(f"verifying {out_path}")
    b3_calc = blake3(open(out_path, "rb").read()).hexdigest()
    if not checksum == b3_calc:
        raise RuntimeError(f"Failed integrity check for {out_path}")
    return out_path



"""*Audio handling module*."""

import shutil
import tempfile
from pathlib import Path
from PIL import Image, ImageEnhance
from loguru import logger as log
import json
import taglib
import iscc_sdk as idk


__all__ = [
    "audio_meta_embed",
    "audio_meta_extract",
    "audio_features_extract",
    "audio_thumbnail",
]


AUDIO_META_MAP = {
    "ISCC:NAME": "name",
    "ISCC:DESCRIPTION": "description",
    "ISCC:META": "meta",
    "ISCC:LICENSE": "license",
    "ISCC:ACQUIRE": "acquire",
    "TITLE": "name",
    # "ALBUM": "album",
    # "GENRE": "genre",
    "COMPOSER": "creator",
    "ORIGINALARTIST": "creator",
    "ARTIST": "creator",
    "ALBUMARTIST": "creator",
    "LICENSE": "license",
    "COPYRIGHT": "rights",
    "LANGUAGE": "language",
    "URL": "acquire",
    "ARTISTWEBPAGE": "acquire",
}


def audio_thumbnail(fp):
    # type: (str|Path) -> Image.Image|None
    """
    Create a thumbnail from embedded cover art.

    :param fp: Filepath to audio file.
    :return: Thumbnail image as PIL Image object
    """
    fp = Path(fp)
    tempdir = Path(tempfile.mkdtemp())
    tempimg = tempdir / "cover.jpg"
    cmd = ["-i", fp, "-an", "-vcodec", "copy", tempimg]
    size = idk.sdk_opts.image_thumbnail_size
    try:
        idk.run_ffmpeg(cmd)
        img = Image.open(tempimg)
    except Exception:
        img = None
    if img:
        img.thumbnail((size, size), resample=idk.LANCZOS)
        img = ImageEnhance.Sharpness(img.convert("RGB")).enhance(1.4)

    shutil.rmtree(tempdir, ignore_errors=True)
    return img


def audio_meta_extract(fp):
    # type: (str|Path) -> dict
    """
    Extract metadata from audio file.

    :param fp: Filepath to audio file.
    :return: Metadata mapped to IsccMeta schema
    """
    fp = Path(fp)
    mapped = dict()
    done = set()

    try:
        obj = taglib.File(fp)
        meta = dict(obj.tags)
        mapped["duration"] = obj.length
        obj.close()
    except OSError as e:  # pragma: no cover
        # This is a workaround for the issue that taglib requires exclusive access even for reading.
        log.warning(f"Create tempfile for taglib access {fp.name}: {e}")
        try:
            with idk.TempFile(fp) as tmp_path:
                obj = taglib.File(tmp_path.as_posix())
                meta = dict(obj.tags)
                mapped["duration"] = obj.length
                obj.close()
        except Exception as e:
            log.warning(f"Failed metadata extraction for {fp.name}: {e}")
            return mapped

    for tag, mapped_field in AUDIO_META_MAP.items():
        if mapped_field in done:
            continue
        value = meta.get(tag)
        if value:
            log.debug(f"Mapping audio metadata: {tag} -> {mapped_field} -> {value[0]}")
            mapped[mapped_field] = value[0]
            done.add(mapped_field)
    # Todo - add bitrate, channels, samplerate to iscc-schema
    # mapped["bitrate"] = obj.bitrate
    # mapped["channels"] = obj.channels
    # mapped["samplerate"] = obj.sampleRate
    return mapped


def audio_meta_embed(fp, meta):
    # type: (str|Path, idk.IsccMeta) -> str
    """
    Embed metadata into a copy of the audio file.

    :param fp: Filepath to source audio file
    :param IsccMeta meta: Metadata to embed into audio file
    :return: Filepath to new audio file with updated metadata
    """
    fp = Path(fp)
    tdir = tempfile.mkdtemp()
    tfile = shutil.copy(fp, tdir)
    obj = taglib.File(tfile)
    if meta.name:
        obj.tags["TITLE"] = [meta.name]
        obj.tags["ISCC:NAME"] = [meta.name]
    if meta.description:
        obj.tags["ISCC:DESCRIPTION"] = [meta.description]
    if meta.meta:
        obj.tags["ISCC:META"] = [meta.meta]
    if meta.license:
        obj.tags["ISCC:LICENSE"] = [meta.license]
    if meta.acquire:
        obj.tags["ISCC:ACQUIRE"] = [meta.acquire]
    obj.save()
    obj.close()
    return tfile


def audio_features_extract(fp):
    # type: (str|Path) -> dict
    """
    Exctracts chromprint fingerprint.

    :param fp: Filepath
    :return: A dict with `duration` in seconds and `fingerprint` 32-bit integers
    """
    fp = Path(fp)
    args = ["-raw", "-json", "-signed", "-length", "0", fp]
    proc = idk.run_fpcalc(args)
    result = json.loads(proc.stdout)
    return result



import tempfile
from pathlib import Path

from docx import Document
import iscc_sdk as idk
import shutil

__all__ = [
    "docx_meta_embed",
]


META_DOCX_MAP = {
    "name": "title",
    "description": "comments",
    "creator": "author",
    "keywords": "keywords",
}


def docx_meta_embed(fp, meta):
    # type: (str|Path, idk.IsccMeta) -> str
    """
    Embed metadata into a copy of the DOCX file.

    :param fp: Filepath to source DOCX file
    :param meta: Metadata to embed into DOCX
    :return: Filepath to the new DOCX file with updated metadata
    """
    fp = Path(fp)
    tempdir = tempfile.mkdtemp()
    tempdoc = shutil.copy(fp, tempdir)
    doc = Document(fp.as_posix())
    new_meta = doc.core_properties
    for iscc_field, docx_field in META_DOCX_MAP.items():
        value = getattr(meta, iscc_field)
        if value:
            setattr(new_meta, docx_field, value)
    doc.save(tempdoc)
    return tempdoc
